'use strict';

const _ = require('lodash');
const utilsFunctions = require('./lib/utils');
const mysqlFunctions = require('./lib/databases/mysql')({
  binDir: $app.binDir, logger: utilsFunctions.getLoggerFromApp($app),
});
const networkFunctions = require('./lib/network')($app);
const componentFunctions = require('./lib/component')($app);

$app.postInstallation = function() {
  const adminConnection = {
    host: $app.host,
    port: $app.port,
    user: $app.rootUser,
    password: $app.rootPassword,
    sslCAFile: $app.sslCAFile,
  };

  if ($app.sslEnable === 'yes') {
    // Wrap binary to force the usage of SSL
    const wrapper = $file.join($app.binDir, 'mysql');
    const wrappedBinary = $file.join($app.installdir, '.bin', 'mysql');
    $file.move(wrapper, wrappedBinary);
    $hb.renderToFile('mysql.sh.tpl', wrapper, {wrappedBinary});
    $file.chmod(wrapper, '755');
  }

  // If all the mandatory parameters are provided, attempt to create the database
  if (![$app.createDatabaseName, $app.createDatabaseUser].some(_.isEmpty)) {
    networkFunctions.waitForService($app.host, $app.port);

    // Wait for a valid root connection
    mysqlFunctions.checkConnection(adminConnection);

    // Check whether or not the database truly exists
    const existsDatabase = mysqlFunctions.existsDatabase($app.createDatabaseName, adminConnection);
    // In case the database does not exist it will be created.
    if (!existsDatabase) {
      // Validate password
      componentFunctions.validatePasswords(
        {password: $app.createDatabasePassword},
        {allowEmptyPassword: $app.allowEmptyPassword}
      );
      $app.info(`==> Creating database ${$app.createDatabaseName}...`);
      mysqlFunctions.createDatabase($app.createDatabaseName, adminConnection);

      if (mysqlFunctions.existsUser($app.createDatabaseUser, adminConnection)) {
        $app.info(`==> Found existing user ${$app.createDatabaseUser}. Skipping user creation...`);
      } else {
        $app.info(`==> Creating user ${$app.createDatabaseUser}...`);
        mysqlFunctions.createUser($app.createDatabaseUser, _.merge({
          newUserPassword: $app.createDatabasePassword,
        }, adminConnection));
      }

      $app.info(`==> Granting access to ${$app.createDatabaseUser} to the database ${$app.createDatabaseName}...`);
      mysqlFunctions.grantPrivileges($app.createDatabaseName, $app.createDatabaseUser, _.merge({
        privileges: $app.createDatabasePrivileges.split(/[;,]+/g).filter((p) => !_.isEmpty(p)),
      }, adminConnection));

      // Flush privileges cache
      mysqlFunctions.flushPrivileges(adminConnection);

      // Print information only when a database is created
      componentFunctions.printProperties({
        'Database User created': $app.createDatabaseUser,
        'Database Name created': $app.createDatabaseName,
        'Database Password created': $app.createDatabasePassword,
        'Database Host': $app.host,
        'Database Port': $app.port,
        'Database Root User': $app.rootUser,
        'Database Root Password': $app.rootPassword,
      });
    } else {
      $app.info(`==> Found database ${$app.createDatabaseName}. Skipping database creation...`);
    }
  }
};

$app.exports = {
  execute: mysqlFunctions.execute,
  executeSqlFile: mysqlFunctions.executeSqlFile,
  set: mysqlFunctions.set,
  get: mysqlFunctions.get,
  createDatabase: mysqlFunctions.createDatabase,
  createUser: mysqlFunctions.createUser,
  dropUser: mysqlFunctions.dropUser,
  dropDatabase: mysqlFunctions.dropDatabase,
  grantPrivileges: mysqlFunctions.grantPrivileges,
  insert: mysqlFunctions.insert,
  existsDatabase: mysqlFunctions.existsDatabase,
  checkConnection: mysqlFunctions.checkConnection,
  canConnect: mysqlFunctions.canConnect,
  showDatabases: mysqlFunctions.showDatabases,
  showUsers: mysqlFunctions.showUsers,
  existsUser: mysqlFunctions.existsUser,
  flushPrivileges: mysqlFunctions.flushPrivileges,
};
